final int DIR_NONE = 0;
final int DIR_UP = 1;
final int DIR_DOWN = 2;
final int DIR_LEFT = 4;
final int DIR_RIGHT = 8;

int playerX = 0;
int playerY = 0;
int playerDirection = DIR_DOWN;

float cellSize = 16;
float xOffset = 0;
float yOffset = 0;

PFont boxFont;
PFont targetFont;

PImage IMG_PLAYER_UP;
PImage IMG_PLAYER_DOWN;
PImage IMG_PLAYER_LEFT;
PImage IMG_PLAYER_RIGHT;
PImage IMG_GROUND;
PImage IMG_WALL;
PImage IMG_TARGET;
PImage IMG_BOX_RIGHT;
PImage IMG_BOX_WRONG;

void setupImages() {
  IMG_PLAYER_UP = loadImage("data/kenney_sokobanpack/player/player_06.png");
  IMG_PLAYER_DOWN = loadImage("data/kenney_sokobanpack/player/player_03.png");
  IMG_PLAYER_LEFT = loadImage("data/kenney_sokobanpack/player/player_18.png");
  IMG_PLAYER_RIGHT = loadImage("data/kenney_sokobanpack/player/player_15.png");
  IMG_GROUND = loadImage("data/kenney_sokobanpack/ground/ground_04.png");
  IMG_WALL = loadImage("data/kenney_sokobanpack/blocks/block_05.png");
  IMG_TARGET = loadImage("data/kenney_sokobanpack/ground/ground_02.png");
  IMG_BOX_RIGHT = loadImage("data/kenney_sokobanpack/crates/crate_10.png");
  IMG_BOX_WRONG = loadImage("data/kenney_sokobanpack/crates/crate_09.png");
}

void drawPlaying() {  
  background(#000000);
  
  drawGrid();  
}

void drawPlayer(float x, float y) {
  PImage image;
  
  switch (playerDirection) {
    default:
    case DIR_UP:
      image = IMG_PLAYER_UP;
      break;
    case DIR_DOWN:
      image = IMG_PLAYER_DOWN;
      break;
    case DIR_LEFT:
      image = IMG_PLAYER_LEFT;
      break;
    case DIR_RIGHT:
      image = IMG_PLAYER_RIGHT;
      break;
  }
  
  imageMode(CENTER);
  image(image, xOffset + x * cellSize + cellSize / 2, yOffset + y * cellSize + cellSize / 2, 4 * cellSize / 5, 4 * cellSize / 5); 
}

void drawObject(char object, float x, float y, boolean fittingHere) {
  if (object < 'a' || object > 'z') return;
  
  PImage image = fittingHere ? IMG_BOX_RIGHT : IMG_BOX_WRONG;
  
  imageMode(CENTER);
  image(image, xOffset + x * cellSize + cellSize / 2, yOffset + y * cellSize + cellSize / 2, 4 * cellSize / 5, 4 * cellSize / 5); 
  
  if (hasMultipleTargetsOrBoxes) {
    textFont(boxFont);
    textAlign(CENTER, CENTER);
    fill(#000000);
    text(object, xOffset + x * cellSize + cellSize / 2, yOffset + y * cellSize + 2 * cellSize / 5);
  }
}

void drawTile(char tile, float x, float y) {
  if (tile != '#' && tile != ' ' && (tile < 'A' || tile > 'Z')) return;
  
  PImage image = null;
  
  if (tile == '#') {
    image = IMG_WALL;
  } else if (tile != ' ') {
    image = IMG_TARGET;
  }
  
  imageMode(CENTER);
  image(IMG_GROUND, xOffset + x * cellSize + cellSize / 2, yOffset + y * cellSize + cellSize / 2, cellSize, cellSize);
  
  if (image != null) {
    image(image, xOffset + x * cellSize + cellSize / 2, yOffset + y * cellSize + cellSize / 2, cellSize, cellSize);
  }
  
  if (hasMultipleTargetsOrBoxes && tile >= 'A' && tile <= 'Z') {
    textFont(targetFont);
    textAlign(CENTER, CENTER);
    fill(#000000);
    
    text(tile, xOffset + x * cellSize + cellSize / 2, yOffset + y * cellSize + 2 * cellSize / 5);
  }
}

void drawGrid() {
  if (columns > rows) {
    cellSize = (width - 20) / columns;
  } else {
    cellSize = (height - 20) / rows;
  }
  
  xOffset = (width - cellSize * columns) / 2;
  yOffset = (height - cellSize * rows) / 2;
  
  boxFont = createFont("data/kenney_fontpackage/Kenney Blocks.ttf", 2 * cellSize / 5);
  targetFont = createFont("data/kenney_fontpackage/Kenney Blocks.ttf", 3 * cellSize / 5);
  
  if (grid == null || grid.length <= 0) return;
  
  for (int y = 0; y < rows; y++) {
    String tileRow = grid[y];
    String objectRow = objects[y];
    
    for (int x = 0; x < columns; x++) {
      char tile = tileRow.charAt(x);
      char object = objectRow.charAt(x);
      
      drawTile(tile, x, y);
      drawObject(object, x, y, ("" + tile).equalsIgnoreCase("" + object));
      
      if (playerX == x && playerY == y) {
        drawPlayer(x, y);
      }
    }
  }
  
  stroke(#222222);
  for (int x = 0; x <= columns; x++) {
    line(xOffset + x * cellSize, yOffset, xOffset + x * cellSize, height - yOffset);
  }
  
  for (int y = 0; y <= rows; y++) {
    line(xOffset, yOffset + y * cellSize, width - xOffset, yOffset + y * cellSize);      
  }
}

void move(int direction) {
  char tile;
  char object;
  char nextTile;
  char nextObject;
  
  playerDirection = direction;
  
  switch (direction) {
    case DIR_UP:
      moves.append("U");
      
      if (playerY <= 0) return;
      
      tile = grid[playerY - 1].charAt(playerX);
      object = objects[playerY - 1].charAt(playerX);
      
      if (object >= 'a' && object <= 'z') {
        if (playerY <= 1) return;
        
        nextTile = grid[playerY - 2].charAt(playerX);
        nextObject = objects[playerY - 2].charAt(playerX);
        
        if (nextTile != '#' && nextObject == ' ') {
          objects[playerY - 1] = objects[playerY - 1].substring(0, playerX) + ' ' + objects[playerY - 1].substring(playerX + 1);
          objects[playerY - 2] = objects[playerY - 2].substring(0, playerX) + object + objects[playerY - 2].substring(playerX + 1);
          playerY--;
        }
      } else if (tile != '#') {
        playerY--;
      }
      break;
    case DIR_DOWN:
      moves.append("D");
      
      if (playerY >= rows) return;
      
      tile = grid[playerY + 1].charAt(playerX);
      object = objects[playerY + 1].charAt(playerX);
      
      if (object >= 'a' && object <= 'z') {
        if (playerY >= rows - 1) return;
        
        nextTile = grid[playerY + 2].charAt(playerX);
        nextObject = objects[playerY + 2].charAt(playerX);
        
        if (nextTile != '#' && nextObject == ' ') {
          objects[playerY + 1] = objects[playerY + 1].substring(0, playerX) + ' ' + objects[playerY + 1].substring(playerX + 1);
          objects[playerY + 2] = objects[playerY + 2].substring(0, playerX) + object + objects[playerY + 2].substring(playerX + 1);
          playerY++;
        }
      } else if (tile != '#') {
        playerY++;
      }
      break;
    case DIR_LEFT:
      moves.append("L");
      
      if (playerX <= 0) return;
      
      tile = grid[playerY].charAt(playerX - 1);
      object = objects[playerY].charAt(playerX - 1);
      
      if (object >= 'a' && object <= 'z') {
        if (playerX <= 1) return;
        
        nextTile = grid[playerY].charAt(playerX - 2);
        nextObject = objects[playerY].charAt(playerX - 2);
        
        if (nextTile != '#' && nextObject == ' ') {
          objects[playerY] = objects[playerY].substring(0, playerX - 1) + ' ' + objects[playerY].substring(playerX - 1 + 1);
          objects[playerY] = objects[playerY].substring(0, playerX - 2) + object + objects[playerY].substring(playerX - 2 + 1);
          playerX--;
        }
      } else if (tile != '#') {
        playerX--;
      }
      break;
    case DIR_RIGHT:
      moves.append("R");
      
      if (playerX >= columns) return;
      
      tile = grid[playerY].charAt(playerX + 1);
      object = objects[playerY].charAt(playerX + 1);
      
      if (object >= 'a' && object <= 'z') {
        if (playerX >= columns - 1) return;
        
        nextTile = grid[playerY].charAt(playerX + 2);
        nextObject = objects[playerY].charAt(playerX + 2);
        
        if (nextTile != '#' && nextObject == ' ') {
          objects[playerY] = objects[playerY].substring(0, playerX + 1) + ' ' + objects[playerY].substring(playerX + 1 + 1);
          objects[playerY] = objects[playerY].substring(0, playerX + 2) + object + objects[playerY].substring(playerX + 2 + 1);
          playerX++;
        }
      } else if (tile != '#') {
        playerX++;
      }
      break;
  }
  
  boolean hasWrongObjects = false;
  
  for (int y = 0; y < rows; y++) {
    for (int x = 0; x < columns; x++) {
      object = objects[y].charAt(x);
      tile = grid[y].charAt(x);
      
      if (object >= 'a' && object <= 'z') {
        if (!("" + object).equalsIgnoreCase("" + tile)) {
          hasWrongObjects = true;
        }
      }
    }
  }
  
  if (!hasWrongObjects) {
    endGame();
  }
}

void keyPressedPlaying() {
  if (key == 'r') {
    startGame(selectedLevel);
  } else if (key == 'l') {
    gameState = GAME_LEVEL_SELECT;
  } else if (keyCode == UP) {
    move(DIR_UP);
  } else if (keyCode == DOWN) {
    move(DIR_DOWN);
  } else if (keyCode == LEFT) {
    move(DIR_LEFT);
  } else if (keyCode == RIGHT) {
    move(DIR_RIGHT);
  }
}

void endGame() {  
  if (selectedLevel >= unlockedLevel) {
    unlockedLevel = selectedLevel + 1;
  }
  
  gameState = GAME_OVER;
}
